from __future__ import annotations

import io
from datetime import datetime, timedelta, timezone
from io import BytesIO
from typing import TYPE_CHECKING

import pytest
from dissect.ntfs.secure import ACL, SecurityDescriptor
from flow.record.fieldtypes import command
from flow.record.fieldtypes import datetime as dt

from dissect.target.helpers.regutil import VirtualHive, VirtualKey
from dissect.target.plugins.os.windows.defender._plugin import MicrosoftDefenderPlugin, parse_iso_datetime
from dissect.target.plugins.os.windows.defender.quarantine import (
    STREAM_ID,
    c_defender,
    rc4_crypt,
    recover_quarantined_file_streams,
)
from tests._utils import absolute_path

if TYPE_CHECKING:
    from pathlib import Path

    from flow.record import Record

    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        ("2025-11-26T22:01:52.403Z", datetime(2025, 11, 26, 22, 1, 52, 403000, tzinfo=timezone.utc)),
        (
            "2025-11-26T22:01:52.403+02:00",
            datetime(2025, 11, 26, 22, 1, 52, 403000, tzinfo=timezone(timedelta(hours=2))),
        ),
        ("2025-11-26T22:01:52.403", datetime(2025, 11, 26, 22, 1, 52, 403000, tzinfo=timezone.utc)),
        ("2025-11-26T22:01:52Z", datetime(2025, 11, 26, 22, 1, 52, tzinfo=timezone.utc)),
        ("2025-11-26T22:01:52+00:00", datetime(2025, 11, 26, 22, 1, 52, tzinfo=timezone.utc)),
    ],
)
def test_parse_iso_datetime(value: str, expected: datetime) -> None:
    assert parse_iso_datetime(value) == expected


def test_defender_evtx_logs(target_win: Target, fs_win: VirtualFilesystem, tmp_path: Path) -> None:
    # map default log location to pass EvtxPlugin's compatibility check
    fs_win.map_dir("windows/system32/winevt/logs", tmp_path)

    log_file = absolute_path("_data/plugins/os/windows/defender/operational.evtx")
    fs_win.map_file("windows/system32/winevt/logs/Microsoft-Windows-Windows Defender%4Operational.evtx", log_file)

    target_win.add_plugin(MicrosoftDefenderPlugin)

    records = list(target_win.defender.evtx())

    assert len(records) == 9

    # verify that all records have unique EventIDs
    assert all(r.ts is not None for r in records)
    assert len([r.EventID for r in records]) == 9
    assert {r.Provider_Name for r in records} == {"Microsoft-Windows-Windows Defender"}
    assert {r.Channel for r in records} == {"Microsoft-Windows-Windows Defender/Operational"}
    # Both informational records (no threat name) and detections are present
    assert {r.Threat_Name for r in records} == {None, "TrojanDropper:PowerShell/PowerSploit.S!MSR"}


def test_defender_quarantine_entries(target_win: Target, fs_win: VirtualFilesystem) -> None:
    quarantine_dir = absolute_path("_data/plugins/os/windows/defender/quarantine")

    fs_win.map_dir("programdata/microsoft/windows defender/quarantine", quarantine_dir)

    target_win.add_plugin(MicrosoftDefenderPlugin)

    records = list(target_win.defender.quarantine())

    assert len(records) == 1

    # Test whether the quarantining of a Mimikatz binary is properly parsed.
    mimikatz_record = records[0]
    detection_date = datetime.strptime("2022-12-02", "%Y-%m-%d").replace(tzinfo=timezone.utc).date()

    assert mimikatz_record.detection_type == "file"
    assert mimikatz_record.detection_name == "HackTool:Win64/Mikatz!dha"
    assert mimikatz_record.detection_path == "C:\\Users\\user\\Downloads\\mimikatz\\mimilib.dll"

    assert mimikatz_record.ts.date() == detection_date
    assert mimikatz_record.creation_time.date() == detection_date
    assert mimikatz_record.last_write_time.date() == detection_date
    assert mimikatz_record.last_accessed_time.date() == detection_date

    assert mimikatz_record.quarantine_id == "a762038000000000fb1112639186e0d6"
    assert mimikatz_record.scan_id == "cdbe4600e43a964b8dc2416b0ef7a207"
    assert mimikatz_record.threat_id == 2147705511


def test_defender_quarantine_recovery(target_win: Target, fs_win: VirtualFilesystem, tmp_path: Path) -> None:
    # Map the quarantine folder from our test data
    quarantine_dir = absolute_path("_data/plugins/os/windows/defender/quarantine")
    fs_win.map_dir("programdata/microsoft/windows defender/quarantine", quarantine_dir)

    # Create a directory to recover to
    recovery_dst = tmp_path.joinpath("recovery")
    recovery_dst.mkdir()

    # Recover
    target_win.add_plugin(MicrosoftDefenderPlugin)
    target_win.defender.recover(output_dir=recovery_dst)

    # Set up variables to indicate what we expect to find
    payload_filename = "A6C8322B8A19AEED96EFBD045206966DA4C9619D"
    security_descriptor_filename = "A6C8322B8A19AEED96EFBD045206966DA4C9619D.security_descriptor"
    zone_identifier_filename = "A6C8322B8A19AEED96EFBD045206966DA4C9619D.ZoneIdentifierDATA"
    expected_zone_identifier_content = (
        b"[ZoneTransfer]\r\nZoneId=3\r\nReferrerUrl=C:\\Users\\user\\Downloads\\mimikatz_trunk.zip\r\n"
    )
    expected_owner = "S-1-5-21-2614236324-1336345114-3023566343-1000"
    expected_group = "S-1-5-21-2614236324-1336345114-3023566343-513"

    expected_files = [payload_filename, security_descriptor_filename, zone_identifier_filename]
    expected_files.sort()

    directory_content = [p.name for p in recovery_dst.iterdir()]
    directory_content.sort()
    assert expected_files == directory_content
    # Replaced the mimikatz payload with `DUMMY_PAYLOAD` to avoid defender collecting it
    assert recovery_dst.joinpath(payload_filename).read_bytes() == b"DUMMY_PAYLOAD"

    # Verify that the security descriptors are valid security descriptors
    descriptor_buf = recovery_dst.joinpath(security_descriptor_filename).read_bytes()
    descriptor = SecurityDescriptor(BytesIO(descriptor_buf))

    assert isinstance(descriptor.dacl, ACL)
    assert isinstance(descriptor.sacl, ACL)

    assert descriptor.owner == expected_owner
    assert descriptor.group == expected_group

    # Verify valid zone identifier for mimikatz file
    assert recovery_dst.joinpath(zone_identifier_filename).read_bytes() == expected_zone_identifier_content


def test_defender_exclusions(target_win: Target, hive_hklm: VirtualHive) -> None:
    # https://learn.microsoft.com/en-us/exchange/antispam-and-antimalware/windows-antivirus-software?view=exchserver-2019
    exclusions_example = {
        "Extensions": [".config", ".log", ".cfg"],
        "Processes": ["UmService.exe", "UmWorkerProcess.exe"],
        "Paths": ["C:\\System32\\Cluster\\"],
        "IpAddresses": [],
        "TemporaryPaths": [],
    }
    # Recreate the 'Exclusions' registry key based on the example dict
    exclusions_key = VirtualKey(hive_hklm, "Software\\Microsoft\\Windows Defender\\Exclusions")
    for exclusion_type, exclusions in exclusions_example.items():
        exclusion_type_key = VirtualKey(hive_hklm, exclusions_key.path + f"\\{exclusion_type}")
        for exclusion in exclusions:
            exclusion_type_key.add_value(exclusion, 0)
        exclusions_key.add_subkey(exclusion_type, exclusion_type_key)

    hive_hklm.map_key(exclusions_key.path, exclusions_key)
    target_win.add_plugin(MicrosoftDefenderPlugin)

    exclusion_records = list(target_win.defender.exclusions())

    # If an exclusion type does not have any exclusions, no records are returned for that type.
    ip_address_exclusions = [exclusion for exclusion in exclusion_records if exclusion.type == "IpAddresses"]
    temporary_path_exclusions = [exclusion for exclusion in exclusion_records if exclusion.type == "TemporaryPaths"]
    assert len(ip_address_exclusions) == 0
    assert len(temporary_path_exclusions) == 0

    extension_exclusions = [exclusion for exclusion in exclusion_records if exclusion.type == "Extensions"]
    assert len(extension_exclusions) == 3

    assert any(exclusion.value == ".config" for exclusion in extension_exclusions)
    assert any(exclusion.value == ".log" for exclusion in extension_exclusions)
    assert any(exclusion.value == ".cfg" for exclusion in extension_exclusions)

    process_exclusions = [exclusion for exclusion in exclusion_records if exclusion.type == "Processes"]
    assert len(process_exclusions) == 2
    assert any(exclusion.value == "UmService.exe" for exclusion in process_exclusions)
    assert any(exclusion.value == "UmWorkerProcess.exe" for exclusion in process_exclusions)

    path_exclusions = [exclusion for exclusion in exclusion_records if exclusion.type == "Paths"]
    assert any(exclusion.value == "C:\\System32\\Cluster\\" for exclusion in path_exclusions)
    assert len(path_exclusions) == 1

    assert len(exclusion_records) == 6


def _mplog_records(target_win: Target, fs_win: VirtualFilesystem, tmp_path: Path, log_filename: str) -> list[Record]:
    # map default log location to pass EvtxPlugin's compatibility check
    fs_win.map_dir("windows/system32/winevt/logs", tmp_path)

    log_file = absolute_path(f"_data/plugins/os/windows/defender/mplog/{log_filename}.log")
    fs_win.map_file("ProgramData/Microsoft/Windows Defender/Support/MPLog-20240101-094808.log", log_file)

    target_win.add_plugin(MicrosoftDefenderPlugin)
    return list(target_win.defender.mplog())


def test_defender_mplogs_rtp(target_win: Target, fs_win: VirtualFilesystem, tmp_path: Path) -> None:
    record = _mplog_records(target_win, fs_win, tmp_path, "rtp").pop()
    assert record.source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert record.ts == dt("2021-05-04 09:54:06+00:00")
    assert record.last_perf == dt("2021-05-04 09:54:06+00:00")
    assert record.first_rtp_scan == dt("2021-05-04 09:54:06+00:00")
    assert record.plugin_states == "AV:2  AS:2  RTP:2  OA:2  BM:2"
    assert record.process_exclusions == ["C:\\ReportingServicesService.exe", "C:\\ReportingServicesService.exe"]
    assert sorted(record.path_exclusions) == sorted(
        ["C:\\Windows\\Security\\Database\\*.chk", "%windidr%\\SoftwareDistribution\\Datastore\\*.*"]
    )
    assert sorted(record.ext_exclusions) == sorted([".DBF", ".NDF", ".RAR", ".XML", ".IDX", ".BAK", ".PDF", ".BKP"])


def test_defender_mplogs_resource_scan(target_win: Target, fs_win: VirtualFilesystem, tmp_path: Path) -> None:
    record = _mplog_records(target_win, fs_win, tmp_path, "resourcescan").pop()
    assert record.source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert record.ts == dt("2023-01-01 00:00:00+00:00")
    assert record.scan_id == "{1A2B3C4D-5E6F-7A8B-9C0D-1E2F3A4B5C6D}"
    assert record.scan_source == 4
    assert record.end_time == dt("2023-01-01 00:01:00+00:00")
    assert record.resource_schema == "webfile"
    assert (
        record.resource_path
        == "C:\\Users\\user\\Downloads\\file.rar|https:/example.com/download?id=12345|browser_broker.exe"
    )
    assert record.result_count == 2
    assert sorted(record.threats) == sorted(["Worm:VBS/Jenxcus.DN", "HackTool:MSIL/Mimikatz!MSR"])
    assert sorted(record.resources) == sorted(
        [
            "C:\\ProgramData\\App\\Scans\\FilesStash\\9F8E7D6C-5A4B-3C2D-1E0F-A1B2C3D4E5F6_1d23456789abcdef|C:\\Users\\user\\Downloads\\file.rar",
            "C:\\Users\\user\\Downloads\\file.rar->file.WsF",
            "C:\\ProgramData\\App\\Scans\\FilesStash\\9F8E7D6C-5A4B-3C2D-1E0F-A1B2C3D4E5F6_1d23456789abcdef",
            "C:\\Users\\user\\Downloads\\file.rar|https:/example.com/download?id=67890|browser_broker.exe",
            "C:\\ProgramData\\App\\Scans\\FilesStash\\9F8E7D6C-5A4B-3C2D-1E0F-A1B2C3D4E5F6_1d23456789abcdef|https:/example.com/download?id=12345|browser_broker.exe",
            "C:\\Users\\user\\Downloads\\file.rar",
            "C:\\Users\\user\\Videos\\binary.exe",
        ]
    )


def test_defender_mplogs_threat_actions(target_win: Target, fs_win: VirtualFilesystem, tmp_path: Path) -> None:
    record = _mplog_records(target_win, fs_win, tmp_path, "threatactions").pop()
    assert record.source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert record.ts == dt("2017-12-25 16:30:45+00:00")
    assert record.threats == ["Worm:Win32/RandomName.X"]
    for item in [
        "\\\\?\\C:\\OS\\Tasks\\RandomTask1.job",
        "\\\\?\\C:\\OS\\system32\\randomfile.tmp->(UPX)",
        "\\\\?\\C:\\OS\\Tasks\\RandomTask1.job",
        "\\\\?\\C:\\OS\\system32\\randomfile.tmp",
        "C:\\OS\\Tasks\\RandomTask1.job",
        "\\\\?\\C:\\OS\\Tasks\\RandomTask1.job",
        "C:\\OS\\system32\\randomfile.tmp->(UPX)",
        "\\\\?\\C:\\OS\\system32\\randomfile.tmp->(UPX)",
        "\\\\?\\C:\\OS\\Tasks\\RandomTask1.job",
    ]:
        assert item in list(record.resources)

    assert record.actions == ["quarantine"]


def test_defender_mplogs_bmtelemetry(target_win: Target, fs_win: VirtualFilesystem, tmp_path: Path) -> None:
    record = _mplog_records(target_win, fs_win, tmp_path, "bmtelemetry").pop()
    assert record.source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert record.ts == dt("2024-07-15 11:45:22+00:00")
    assert record.guid == "{1D3E4F07-89AB-45C2-923D-E5F6789A1B2C}"
    assert record.signature_id == 123456789012345
    assert record.sigsha == "abcd1234ef567890abcd1234ef567890abcd1234"
    assert record.threat_level == 0
    assert record.process_id == 8453
    assert record.process_creation_time == 153846236598765432
    assert record.image_path == "C:\\OS\\System32\\servicehost.exe"
    assert record.taint_info == "Friendly: Y; Reason: ; Modules: ; Parents: "
    assert record.operations == "None"


def test_defender_mplogs_lines(target_win: Target, fs_win: VirtualFilesystem, tmp_path: Path) -> None:
    records = _mplog_records(target_win, fs_win, tmp_path, "lines")
    assert len(records) == 10

    # Process Image
    assert records[0].source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert records[0].ts == dt("2024-07-13 14:42:19.659000+00:00")
    assert records[0].process_image_name == "randomapp.exe"
    assert records[0].pid == 5832
    assert records[0].total_time == 1398
    assert records[0].count == 22
    assert records[0].max_time == 398
    assert records[0].max_time_file == "\\Device\\HarddiskVolume2\\Users\\user123\\AppData\\Local\\Temp\\TEMP001.tmp"
    assert records[0].estimated_impact == 4

    # Lowfi
    assert records[1].source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert records[1].ts == dt("2023-01-20 08:45:40.321000+00:00")
    assert records[1].lowfi == command(
        "C:\\OS\\System32\\cfg.exe(reg add HKLM\\SYSTEM\\OtherControlSet\\Control\\SecurityOptions\\SecurityModule /v RandomFlag /t REG_DWORD /d 0 /f)",  # noqa: E501
    )

    # Detection Add
    assert records[2].source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert records[2].ts == dt("2023-01-27 15:33:07.698000+00:00")
    assert (
        records[2].detection
        == "HackTool:MSIL/RndGen!MD5 file:C:\\Users\\user987\\Documents\\executable.exe PropBag [length: 0, data: (null)]"  # noqa: E501
    )

    # Threat
    assert records[3].source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert records[3].ts == dt("2023-01-27 15:33:07.698000+00:00")
    assert records[3].threat == command("C:\\Users\\user987\\Documents\\executable.exe")

    # Detection event
    assert records[4].source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert records[4].ts == dt("2023-01-27 15:33:07.698000+00:00")
    assert records[4].threat_type == "MSIL/RndGen!MD5"
    assert records[4].command == command("C:\\Users\\user987\\Documents\\executable.exe")

    # Exclusion
    assert records[5].source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert records[5].ts == dt("2024-08-17 17:35:22.614000+00:00")
    assert records[5].full_path_with_drive_letter == "C:\\example.txt"
    assert records[5].full_path_with_device_path == "example.txt"

    # Mini-filter unsuccesful scan
    assert records[6].source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert (
        records[6].path
        == "\\Device\\HarddiskVolume2\\Users\\userdefault\\AppData\\Local\\Packages\\MicrosoftBrowser.Default_cw5n8h2txyuma\\LocalState\\ANWebView\\Default\\Popular URLs."  # noqa: E501
    )
    assert records[6].ts == dt("2024-07-13 14:38:15.272000+00:00")
    assert records[6].process == "(unknown)"
    assert records[6].status == "0xc000004a"
    assert records[6].state == "0"
    assert records[6].scan_request == "#16891"
    assert records[6].file_id == "0x200000002c4b5"
    assert records[6].reason == "OnAccess"
    assert records[6].io_status_block_for_new_file == "0x3"
    assert records[6].desired_access == "0x0"
    assert records[6].file_attributes == "0x20"
    assert records[6].scan_attributes == "0x10"
    assert records[6].access_state_flags == "0x802"
    assert records[6].backing_file_info == "0x0, 0x0, 0x0:0\\0x0:0"

    # Mini-filter blocked file
    assert records[7].source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert records[7].ts == dt("2024-07-13 14:38:15.272000+00:00")
    assert (
        records[7].blocked_file
        == "\\Device\\HarddiskVolume2\\Users\\userdefault\\AppData\\Local\\Packages\\MicrosoftBrowser.Default_cw5n8h2txyuma\\LocalState\\ANWebView\\Default\\Popular URLs."  # noqa: E501
    )
    assert records[7].process == "(unknown)"
    assert records[7].status == "0xc000004a"
    assert records[7].state == "0"
    assert records[7].scan_request == "#16891"
    assert records[7].file_id == "0x200000002c4b5"
    assert records[7].reason == "OnAccess"
    assert records[7].io_status_block_for_new_file == "0x3"
    assert records[7].desired_access == "0x0"
    assert records[7].file_attributes == "0x20"
    assert records[7].scan_attributes == "0x10"
    assert records[7].access_state_flags == "0x802"
    assert records[7].backing_file_info == "0x0, 0x0, 0x0:0\\0x0:0"

    # EMS
    assert records[8].source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert records[8].ts == dt("2024-09-05 10:21:39.417000+00:00")
    assert records[8].process == "sysproc"
    assert records[8].pid == 2820
    assert records[8].sigseq == "0x1"
    assert records[8].send_memory_scan_report == 1
    assert records[8].source_id == 4

    # Original Filename
    assert records[9].source_log == "sysvol/programdata/microsoft/windows defender/support/MPLog-20240101-094808.log"
    assert records[9].ts == dt("2024-09-03 18:12:05.364000+00:00")
    assert records[9].original_file_name == "RandomData0123_static.dll"
    assert (
        records[9].full_path
        == "c:\\os\\winsxs\\x64_default-app-service_31bf3856ad364e35_10.0.29999.9999_none_fakef123456789a\\randomdata0123.dll"  # noqa: E501
    )

    assert records[9].hr == "0x1"


@pytest.mark.parametrize(
    ("stream_id", "extension"),
    [
        (
            STREAM_ID.EA_DATA,
            "ea_data",
        ),
        (
            STREAM_ID.SECURITY_DATA,
            "security_descriptor",
        ),
        (
            STREAM_ID.LINK,
            "link",
        ),
        (
            STREAM_ID.PROPERTY_DATA,
            "property_data",
        ),
        (
            STREAM_ID.OBJECT_ID,
            "object_id",
        ),
        (
            STREAM_ID.REPARSE_DATA,
            "reparse_data",
        ),
        (
            STREAM_ID.SPARSE_BLOCK,
            "sparse_block",
        ),
        (
            STREAM_ID.TXFS_DATA,
            "txfs_data",
        ),
        (
            STREAM_ID.GHOSTED_FILE_EXTENTS,
            "ghosted_file_extents",
        ),
    ],
)
def test_recover_quarantined_file_streams_valid_param(stream_id: int, extension: str) -> None:
    quarantine_buf = b"\x69" * 32
    quarantine_stream = rc4_crypt(
        c_defender.WIN32_STREAM_ID(
            stream_id, c_defender.STREAM_ATTRIBUTES.STREAM_NORMAL_ATTRIBUTE, len(quarantine_buf), 0
        ).dumps()
        + quarantine_buf
    )

    filename, filebuf = next(recover_quarantined_file_streams(io.BytesIO(quarantine_stream), "valid_stream"))

    assert filename.endswith(extension)
    assert filebuf == quarantine_buf


def test_recover_quarantined_file_streams_invalid() -> None:
    quarantine_buf = b"\x69" * 32
    valid_quarantine_stream = c_defender.WIN32_STREAM_ID(
        STREAM_ID.EA_DATA, c_defender.STREAM_ATTRIBUTES.STREAM_NORMAL_ATTRIBUTE, len(quarantine_buf), 0
    ).dumps()

    # Make stream ID invalid.
    invalid_quarantine_stream = b"\x69" + valid_quarantine_stream[1:]

    quarantine_stream = rc4_crypt(invalid_quarantine_stream + quarantine_buf)

    with pytest.raises(ValueError, match="Unexpected Stream ID "):
        next(recover_quarantined_file_streams(io.BytesIO(quarantine_stream), "invalid_stream"))


def test_recover_quarantined_file_streams(target_win: Target, fs_win: VirtualFilesystem, tmp_path: Path) -> None:
    quarantine_file = absolute_path(
        "_data/plugins/os/windows/defender/quarantine/ResourceData/A6/A6C8322B8A19AEED96EFBD045206966DA4C9619D"
    )

    with quarantine_file.open("rb") as fh:
        assert list(recover_quarantined_file_streams(fh, quarantine_file.name)) == [
            (
                "A6C8322B8A19AEED96EFBD045206966DA4C9619D.security_descriptor",
                b"\x01\x00\x14\x88\x14\x00\x00\x000\x00\x00\x00\xa4\x00\x00\x00L\x00\x00\x00\x01\x05\x00\x00\x00\x00\x00\x05\x15\x00\x00\x00\xa4\x14\xd2\x9b\x1a\x02\xa7O\x07\xf67\xb4\xe8\x03\x00\x00\x01\x05\x00\x00\x00\x00\x00\x05\x15\x00\x00\x00\xa4\x14\xd2\x9b\x1a\x02\xa7O\x07\xf67\xb4\x01\x02\x00\x00\x02\x00X\x00\x03\x00\x00\x00\x00\x00\x14\x00\xff\x01\x1f\x00\x01\x01\x00\x00\x00\x00\x00\x05\x12\x00\x00\x00\x00\x00\x18\x00\xff\x01\x1f\x00\x01\x02\x00\x00\x00\x00\x00\x05 \x00\x00\x00 \x02\x00\x00\x00\x00$\x00\xff\x01\x1f\x00\x01\x05\x00\x00\x00\x00\x00\x05\x15\x00\x00\x00\xa4\x14\xd2\x9b\x1a\x02\xa7O\x07\xf67\xb4\xe8\x03\x00\x00\x02\x00L\x00\x01\x00\x00\x00\x12\x10D\x00\x00\x00\x00\x00\x01\x01\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x14\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00(\x00\x00\x00I\x00M\x00A\x00G\x00E\x00L\x00O\x00A\x00D\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00",  # noqa: E501
            ),
            ("A6C8322B8A19AEED96EFBD045206966DA4C9619D", b"DUMMY_PAYLOAD"),
            (
                "A6C8322B8A19AEED96EFBD045206966DA4C9619D.ZoneIdentifierDATA",
                b"[ZoneTransfer]\r\nZoneId=3\r\nReferrerUrl=C:\\Users\\user\\Downloads\\mimikatz_trunk.zip\r\n",
            ),
        ]


@pytest.mark.parametrize(
    "path",
    [
        "Windows\\Temp\\MpCmdRun.log",
        "Users\\John\\AppData\\Local\\Temp\\MpCmdRun.log.bak",
    ],
)
def test_defender_mpcmdrunlog(target_win_users: Target, fs_win: VirtualFilesystem, path: str) -> None:
    """Test if we can parse a Windows Defender ``MpCmdRun.log`` file."""

    fs_win.map_file(path, absolute_path("_data/plugins/os/windows/defender/mpcmdrun/MpCmdRun.log"))

    target_win_users.add_plugin(MicrosoftDefenderPlugin)
    records = list(target_win_users.defender.mpcmdrun())

    assert len(records) == 4

    assert path in str(records[0].source)
    assert records[0].ts_start == dt("2025-11-03 13:08:19+00:00")
    assert records[0].ts_end == dt("2025-11-03 13:08:20+00:00")
    assert records[0].command == (
        '"C:\\ProgramData\\Microsoft\\Windows Defender\\platform\\4.18.2203.5-0\\MpCmdRun.exe" -EnableService'
    )

    assert path in str(records[1].source)
    assert records[1].ts_start == dt("2025-11-03 13:09:45+00:00")
    assert records[1].command == (
        '"C:\\ProgramData\\Microsoft\\Windows Defender\\platform\\4.18.2203.5-0\\MpCmdRun.exe" -EnableService'
    )

    assert path in str(records[2].source)
    assert records[2].ts_start == dt("2025-11-03 13:14:55+00:00")
    assert records[2].command == (
        '"C:\\ProgramData\\Microsoft\\Windows Defender\\platform\\4.18.2203.5-0\\MpCmdRun.exe"'
        " SignatureUpdate -ScheduleJob -RestrictPrivileges"
    )

    assert path in str(records[3].source)
    assert records[3].ts_start is None
    assert records[3].command == (
        '"C:\\ProgramData\\Microsoft\\Windows Defender\\platform\\4.18.2203.5-0\\MpCmdRun.exe"'
        " SignaturesUpdateService -ScheduleJob -UncDownload"
    )



from __future__ import annotations

import textwrap
from io import BytesIO, StringIO
from typing import TYPE_CHECKING

import pytest

from dissect.target.helpers.fsutil import stat_result
from dissect.target.tools.diff import (
    DifferentialCli,
    TargetComparison,
    differentiate_target_filesystems,
    differentiate_target_plugin_outputs,
    likely_unchanged,
)
from dissect.target.tools.diff import main as target_diff
from dissect.target.tools.utils import fs
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.target import Target
    from tests.conftest import TargetUnixFactory

PASSWD_CONTENTS = """
            root:x:0:0:root:/root:/bin/bash
            user:x:1000:1000:user:/home/user:/bin/bash
            """


@pytest.fixture
def src_target(target_unix_factory: TargetUnixFactory) -> Target:
    target, fs_unix = target_unix_factory.new("src_target")

    passwd_contents = PASSWD_CONTENTS + "\nsrc_user:x:1001:1001:src_user:/home/src_user:/bin/bash"

    fs_unix.map_file_fh("/etc/passwd", BytesIO(textwrap.dedent(passwd_contents).encode()))

    fs_unix.map_file_fh("changes/unchanged", BytesIO(b"Unchanged"))
    fs_unix.map_file_fh("changes/changed", BytesIO(b"Hello From Source Target"))
    fs_unix.map_file_fh("changes/only_on_src", BytesIO(b"FooBarBaz"))

    fs_unix.map_file_fh("changes/subdirectory_both/on_both", BytesIO(b"On Both"))
    fs_unix.map_file_fh("changes/subdirectory_src/only_on_src", BytesIO(b"Hello From Source Target"))

    fs_unix.map_file_fh("changes/file_on_src", BytesIO(b"Hello From Source Target"))
    fs_unix.map_file_fh("changes/dir_on_src/file", BytesIO(b"Hello From Source Target"))
    return target


@pytest.fixture
def dst_target(target_unix_factory: TargetUnixFactory) -> Target:
    target, fs_unix = target_unix_factory.new("dst_target")

    passwd_contents = PASSWD_CONTENTS + "\ndst_user:x:1002:1002:dst_user:/home/dst_user:/bin/bash"

    fs_unix.map_file_fh("/etc/passwd", BytesIO(textwrap.dedent(passwd_contents).encode()))

    fs_unix.map_file_fh("changes/unchanged", BytesIO(b"Unchanged"))
    fs_unix.map_file_fh("changes/changed", BytesIO(b"Hello From Destination Target"))
    fs_unix.map_file_fh("changes/only_on_dst", BytesIO(b"BazBarFoo"))

    fs_unix.map_file_fh("changes/subdirectory_both/on_both", BytesIO(b"On Both"))
    fs_unix.map_file_fh("changes/subdirectory_dst/only_on_dst", BytesIO(b"Hello From Destination Target"))

    fs_unix.map_file_fh("changes/dir_on_src", BytesIO(b"Hello From Destination Target"))
    fs_unix.map_file_fh("changes/file_on_src/file", BytesIO(b"Hello From Destination Target"))
    return target


def test_scandir(src_target: Target, dst_target: Target) -> None:
    comparison = TargetComparison(src_target, dst_target, deep=True)
    diff = comparison.scandir("changes")

    assert len(diff.deleted) == 4
    assert diff.deleted[0].name == "only_on_src"
    assert diff.deleted[0].open().read() == b"FooBarBaz"
    assert diff.deleted[1].name == "subdirectory_src"
    assert diff.deleted[2].name == "dir_on_src"
    assert diff.deleted[3].open().read() == b"Hello From Source Target"

    assert len(diff.created) == 4
    assert diff.created[0].open().read() == b"BazBarFoo"
    assert diff.created[0].name == "only_on_dst"
    assert diff.created[1].name == "subdirectory_dst"

    assert diff.created[2].name == "dir_on_src"
    assert diff.created[2].open().read() == b"Hello From Destination Target"
    assert diff.created[3].name == "file_on_src"
    assert diff.created[3].is_dir()

    assert len(diff.unchanged) == 2
    assert diff.unchanged[0].open().read() == b"Unchanged"
    assert diff.unchanged[0].name == "unchanged"

    assert diff.unchanged[1].name == "subdirectory_both"

    assert len(diff.modified) == 1
    differential_entry = diff.modified[0]
    assert differential_entry.src_target_entry.open().read() == b"Hello From Source Target"
    assert differential_entry.dst_target_entry.open().read() == b"Hello From Destination Target"
    assert differential_entry.diff == [
        b"--- \n",
        b"+++ \n",
        b"@@ -1 +1 @@\n",
        b"-Hello From Source Target",
        b"+Hello From Destination Target",
    ]


def test_walkdir(src_target: Target, dst_target: Target) -> None:
    comparison = TargetComparison(src_target, dst_target, deep=True)
    differentials = list(comparison.walkdir("changes"))

    assert len(differentials) == 6
    assert sorted(differential.directory for differential in differentials) == [
        "/changes/dir_on_src",
        "/changes/file_on_src",
        "/changes/subdirectory_both",
        "/changes/subdirectory_dst",
        "/changes/subdirectory_src",
        "changes",
    ]

    assert differentials[0].directory == "changes"

    subdirectories_only_on_dst = ["/changes/subdirectory_dst", "/changes/file_on_src"]
    for subdirectory in subdirectories_only_on_dst:
        differential = next(differential for differential in differentials if differential.directory == subdirectory)

        # All entries should be 'created' as this directory doesn't exist on the source target
        assert len(differential.modified) == 0
        assert len(differential.deleted) == 0
        assert len(differential.unchanged) == 0
        assert len(differential.created) == 1
        assert differential.created[0].open().read() == b"Hello From Destination Target"

    subdirectories_only_on_src = ["/changes/subdirectory_src", "/changes/dir_on_src"]

    for subdirectory in subdirectories_only_on_src:
        differential = next(differential for differential in differentials if differential.directory == subdirectory)

        # All entries should be 'created' as this directory doesn't exist on the destination target
        assert len(differential.modified) == 0
        assert len(differential.deleted) == 1
        assert len(differential.unchanged) == 0
        assert len(differential.created) == 0
        assert differential.deleted[0].open().read() == b"Hello From Source Target"

    # All entries should be 'unchanged' as this folder is identical on both
    assert len(differentials[3].modified) == 0
    assert len(differentials[3].deleted) == 0
    assert len(differentials[3].unchanged) == 1
    assert len(differentials[3].created) == 0
    assert differentials[3].unchanged[0].open().read() == b"On Both"


def test_likely_unchanged() -> None:
    # ['mode', 'addr', 'dev', 'nlink', 'uid', 'gid', 'size', 'atime', 'mtime', 'ctime']
    mock_stat = stat_result([0o1777, 1, 2, 3, 1337, 7331, 999, 0, 0, 0])
    mock_stat_accessed = stat_result([0o1777, 1, 2, 3, 1337, 7331, 999, 999, 0, 0])
    mock_stat_changed = stat_result([0o1777, 1, 2, 3, 1337, 7331, 999, 999, 999, 0])

    assert likely_unchanged(mock_stat, mock_stat_accessed)
    assert not likely_unchanged(mock_stat, mock_stat_changed)


def test_differentiate_filesystems(src_target: Target, dst_target: Target) -> None:
    records = list(differentiate_target_filesystems(src_target, dst_target, deep=True, exclude="/etc/*"))

    created = [record for record in records if "created" in record._desc.name]
    modified = [record for record in records if "modified" in record._desc.name]
    deleted = [record for record in records if "deleted" in record._desc.name]

    assert len(created) == 6
    assert all(record._desc.name == "differential/file/created" for record in created)

    assert len(modified) == 1
    assert all(record._desc.name == "differential/file/modified" for record in modified)

    assert len(deleted) == 6
    assert all(record._desc.name == "differential/file/deleted" for record in deleted)


def test_differentiate_plugins(src_target: Target, dst_target: Target) -> None:
    records = list(differentiate_target_plugin_outputs(src_target, dst_target, plugin="users"))
    assert len(records) == 4

    created = [record for record in records if "created" in record._desc.name]
    unchanged = [record for record in records if "unchanged" in record._desc.name]
    deleted = [record for record in records if "deleted" in record._desc.name]

    assert len(unchanged) == 2
    assert len(created) == 1
    assert len(deleted) == 1

    assert created[0].record.name == "dst_user"
    assert created[0].record.hostname == "dst_target"
    assert deleted[0].record.name == "src_user"
    assert deleted[0].record.hostname == "src_target"


def test_shell_ls(
    src_target: Target, dst_target: Target, capsys: pytest.CaptureFixture, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr(fs, "LS_COLORS", {})

    cli = DifferentialCli(src_target, dst_target, deep=True)
    cli.onecmd("ls changes")

    captured = capsys.readouterr()

    expected = [
        "changed (modified)",
        "dir_on_src (created)",
        "dir_on_src (deleted)",
        "file_on_src (created)",
        "file_on_src (deleted)",
        "only_on_dst (created)",
        "only_on_src (deleted)",
        "subdirectory_both",
        "subdirectory_dst (created)",
        "subdirectory_src (deleted)",
        "unchanged",
    ]

    assert captured.out == "\n".join(expected) + "\n"


def test_shell_find(
    src_target: Target, dst_target: Target, capsys: pytest.CaptureFixture, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr(fs, "LS_COLORS", {})

    cli = DifferentialCli(src_target, dst_target, deep=True)
    cli.onecmd("find /changes -cmd")

    captured = capsys.readouterr()

    expected = [
        "/changes/changed (modified)",
        "/changes/dir_on_src (created)",
        "/changes/dir_on_src (deleted)",
        "/changes/file_on_src (created)",
        "/changes/file_on_src (deleted)",
        "/changes/only_on_dst (created)",
        "/changes/only_on_src (deleted)",
        "/changes/subdirectory_dst (created)",
        "/changes/subdirectory_src (deleted)",
        "/changes/subdirectory_dst/only_on_dst (created)",
        "/changes/file_on_src/file (created)",
        "/changes/subdirectory_src/only_on_src (deleted)",
        "/changes/dir_on_src/file (deleted)",
    ]

    assert captured.out == "\n".join(expected) + "\n"


def test_shell_cat(src_target: Target, dst_target: Target, capsys: pytest.CaptureFixture) -> None:
    cli = DifferentialCli(src_target, dst_target, deep=True)

    cli.onecmd("cat /changes/unchanged")
    captured = capsys.readouterr()
    assert captured.out == "Unchanged\n"

    cli.onecmd("cat /changes/subdirectory_dst/only_on_dst")
    captured = capsys.readouterr()
    assert captured.out == "Hello From Destination Target\n"

    cli.onecmd("cat /changes/subdirectory_src/only_on_src")
    captured = capsys.readouterr()
    assert captured.out == "Hello From Source Target\n"

    # When a file is present on both, we want the last version of the file to be outputted.
    cli.onecmd("cat /changes/changed")
    captured = capsys.readouterr()
    assert captured.out == "Hello From Destination Target\n"


def test_shell_plugin(src_target: Target, dst_target: Target, capsys: pytest.CaptureFixture) -> None:
    cli = DifferentialCli(src_target, dst_target, deep=True)

    cli.onecmd("plugin users")
    captured = capsys.readouterr()

    assert "differential/record/created" in captured.out
    assert "differential/record/unchanged" in captured.out
    assert "differential/record/deleted" in captured.out


def test_target_diff_shell(capsys: pytest.CaptureFixture, monkeypatch: pytest.MonkeyPatch) -> None:
    with monkeypatch.context() as m:
        m.setattr(fs, "LS_COLORS", {})
        m.setenv("NO_COLOR", "1")
        src_target_path = str(absolute_path("_data/tools/diff/src.tar"))
        dst_target_path = str(absolute_path("_data/tools/diff/dst.tar"))
        m.setattr("sys.argv", ["target-diff", "--deep", "shell", src_target_path, dst_target_path])
        m.setattr("sys.stdin", StringIO("ls changes"))
        target_diff()
        out, err = capsys.readouterr()
        out = out.replace("(diff) src_target/dst_target:/$", "").strip()

        expected = [
            "changed (modified)",
            "only_on_dst (created)",
            "only_on_src (deleted)",
            "subdirectory_both",
            "subdirectory_dst (created)",
            "subdirectory_src (deleted)",
            "unchanged",
        ]

        assert out == "\n".join(expected)
        assert "unrecognized arguments" not in err


def test_target_diff_fs(capsys: pytest.CaptureFixture, monkeypatch: pytest.MonkeyPatch) -> None:
    with monkeypatch.context() as m:
        src_target_path = str(absolute_path("_data/tools/diff/src.tar"))
        dst_target_path = str(absolute_path("_data/tools/diff/dst.tar"))
        m.setattr("sys.argv", ["target-diff", "--deep", "fs", "--strings", src_target_path, dst_target_path])
        target_diff()
        out, _ = capsys.readouterr()

        assert "differential/file/created" in out
        assert "differential/file/modified" in out
        assert "differential/file/deleted" in out


def test_target_diff_query(capsys: pytest.CaptureFixture, monkeypatch: pytest.MonkeyPatch) -> None:
    with monkeypatch.context() as m:
        src_target_path = str(absolute_path("_data/tools/diff/src.tar"))
        dst_target_path = str(absolute_path("_data/tools/diff/dst.tar"))
        m.setattr("sys.argv", ["target-diff", "query", "--strings", "-f", "users", src_target_path, dst_target_path])
        target_diff()
        out, _ = capsys.readouterr()

        assert "differential/record/created" in out
        assert "differential/record/unchanged" in out
        assert "differential/record/deleted" in out


def test_target_diff_fs_reverse_read(target_unix_factory: TargetUnixFactory) -> None:
    """Test if we detect the difference in an appended file correctly."""

    src_target, fs_src = target_unix_factory.new("src_target")
    dst_target, fs_dst = target_unix_factory.new("dst_target")
    fs_src.map_file_fh("var/log/example.log", BytesIO(b"A" * 1024 * 20))
    fs_dst.map_file_fh("var/log/example.log", BytesIO(b"A" * 1024 * 20 + b"B" * 1024))

    comparison = TargetComparison(src_target, dst_target, deep=True)
    diff = comparison.scandir("/var/log")

    assert len(diff.modified) == 1
    assert diff.modified[0].path == "/var/log/example.log"
    assert diff.modified[0].diff == [
        b"--- \n",
        b"+++ \n",
        b"@@ -1 +1 @@\n",
        b"-" + b"A" * 10 * 1024,
        b"+" + (b"A" * 9216) + (b"B" * 1024),
    ]



from __future__ import annotations

import zipfile
from typing import TYPE_CHECKING
from urllib.parse import quote, unquote

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.filesystems.dir import DirectoryFilesystem
from dissect.target.filesystems.zip import ZipFilesystem
from dissect.target.helpers.fsutil import basename, dirname, join
from dissect.target.helpers.logging import get_logger
from dissect.target.loader import Loader
from dissect.target.loaders.dir import find_dirs, map_dirs
from dissect.target.plugin import OperatingSystem
from dissect.target.plugins.apps.edr.velociraptor import VELOCIRAPTOR_RESULTS

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import Target


log = get_logger(__name__)

FILESYSTEMS_ROOT = "uploads"
UNIX_ACCESSORS = ["file", "auto"]
WINDOWS_ACCESSORS = ["mft", "ntfs", "lazy_ntfs", "ntfs_vss", "auto"]


def find_fs_directories(path: Path) -> tuple[OperatingSystem | None, list[Path] | None]:
    fs_root = path.joinpath(FILESYSTEMS_ROOT)

    # Unix
    for accessor in UNIX_ACCESSORS:
        accessor_root = fs_root.joinpath(accessor)
        if accessor_root.exists():
            os_type, dirs = find_dirs(accessor_root)
            if os_type in [
                OperatingSystem.UNIX,
                OperatingSystem.LINUX,
                OperatingSystem.OSX,
            ]:
                return os_type, [dirs[0]]

    # Windows
    volumes = set()
    vss_volumes = set()
    for accessor in WINDOWS_ACCESSORS:
        accessor_root = fs_root.joinpath(accessor)
        if accessor_root.exists():
            # If the accessor directory exists, assume all the subdirectories are volumes
            for volume in accessor_root.iterdir():
                if not volume.is_dir():
                    continue

                # https://github.com/Velocidex/velociraptor/blob/87368e7cc678144592a1614bb3bbd0a0f900ded9/accessors/ntfs/vss.go#L82
                if "HarddiskVolumeShadowCopy" in volume.name:
                    vss_volumes.add(volume)
                elif (drive_letter := extract_drive_letter(volume.name)) is not None:
                    volumes.add((drive_letter, volume))
                else:
                    volumes.add(volume)

    if volumes:
        # The volumes that represent drives (C, D) are mounted first,
        # otherwise one of the volume shadow copies could be detected as the root filesystem which results in errors.
        return OperatingSystem.WINDOWS, list(volumes) + list(vss_volumes)

    return None, None


def extract_drive_letter(name: str) -> str | None:
    # \\.\X: in URL encoding
    if len(name) == 14 and name.startswith("%5C%5C.%5C") and name.endswith("%3A"):
        return name[10].lower()

    # X: in URL encoding
    if len(name) == 4 and name.endswith("%3A"):
        return name[0].lower()
    return None


class VelociraptorLoader(Loader):
    """Load Rapid7 Velociraptor forensic image files.

    As of Velociraptor version 0.7.0 the structure of the Velociraptor Offline Collector varies by operating system.
    Generic.Collectors.File (Unix) uses the accessors file and auto. The loader supports the following configuration::

        {"Generic.Collectors.File": {"Root": "/", "collectionSpec": "Glob\\netc/**\\nvar/log/**"}}

    Generic.Collectors.File (Windows) and Windows.KapeFiles.Targets (Windows) uses the accessors mft, ntfs, lazy_ntfs,
    ntfs_vss and auto. The loader supports a collection where multiple accessors were used.

    References:
        - https://www.rapid7.com/products/velociraptor/
        - https://docs.velociraptor.app/
        - https://github.com/Velocidex/velociraptor
    """

    def __init__(self, path: Path, **kwargs):
        super().__init__(path, **kwargs)

        if path.suffix == ".zip":
            self.root = zipfile.Path(path.open("rb"))
            if self.root.root.getinfo("uploads.json").compress_type > 0:
                log.warning(
                    "Velociraptor target '%s' is compressed, which will slightly affect performance. "
                    "Consider uncompressing the archive and passing the uncompressed folder to Dissect.",
                    path,
                )
        else:
            self.root = path

    @staticmethod
    def detect(path: Path) -> bool:
        # The 'uploads' folder contains the data acquired
        # The 'results' folder contains information about the used Velociraptor artifacts e.g. Generic.Collectors.File
        # The 'uploads.json' file contains information about the collected files
        # Collection-HOSTNAME-TIMESTAMP/
        #   uploads/
        #   results/
        #   uploads.json
        #   [...] other files related to the collection
        if path.exists() and path.suffix == ".zip":  # novermin
            path = zipfile.Path(path.open("rb"))

        if path.joinpath(FILESYSTEMS_ROOT).exists() and path.joinpath("uploads.json").exists():
            _, dirs = find_fs_directories(path)
            return bool(dirs)

        return False

    def map(self, target: Target) -> None:
        os_type, dirs = find_fs_directories(self.root)

        # Velociraptor URL encodes paths before storing these in a collection, this leads plugins not being able to find
        # these paths. To circumvent this issue, for a zip file the path names are URL decoded before mapping into the
        # VFS and for a directory the paths are URL encoded at lookup time.
        map_dirs(
            target,
            dirs,
            os_type,
            dirfs=VelociraptorDirectoryFilesystem,
            zipfs=VelociraptorZipFilesystem,
        )

        if (results := self.root.joinpath("results")).is_dir():
            # Map artifact results collected by Velociraptor
            vfs = VirtualFilesystem()

            for artifact in results.iterdir():
                if not artifact.name.endswith(".json"):
                    continue

                vfs.map_file_fh(artifact.name, artifact.open("rb"))

            if (uploads := self.root.joinpath("uploads.json")).exists():
                vfs.map_file_fh(uploads.name, uploads.open("rb"))

            target.fs.mount(VELOCIRAPTOR_RESULTS, vfs)


class VelociraptorDirectoryFilesystem(DirectoryFilesystem):
    def _resolve_path(self, path: str) -> Path:
        path = quote(path, safe="$/% ")
        if (fname := basename(path)).startswith("."):
            path = join(dirname(path), fname.replace(".", "%2E", 1))

        return super()._resolve_path(path)


class VelociraptorZipFilesystem(ZipFilesystem):
    def _resolve_path(self, path: str) -> str:
        return unquote(super()._resolve_path(path))



from __future__ import annotations

import io
import uuid
from datetime import timezone
from typing import TYPE_CHECKING, Any, Final

from dissect.cstruct import cstruct
from dissect.util.ts import dostimestamp
from flow.record.fieldtypes import windows_path

from dissect.target.exceptions import RegistryKeyNotFoundError, UnsupportedPluginError
from dissect.target.helpers import shell_folder_ids
from dissect.target.helpers.descriptor_extensions import (
    RegistryRecordDescriptorExtension,
    UserRecordDescriptorExtension,
)
from dissect.target.helpers.logging import get_logger
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime

    from dissect.target.helpers.regutil import RegistryKey
    from dissect.target.target import Target


log = get_logger(__name__)

bag_def = """
enum ROOTFOLDER_ID : uint8 {
    INTERNET_EXPLORER   = 0x00,
    LIBRARIES           = 0x42,
    USERS               = 0x44,
    MY_DOCUMENTS        = 0x48,
    MY_COMPUTER         = 0x50,
    NETWORK             = 0x58,
    RECYCLE_BIN         = 0x60,
    INTERNET_EXPLORER   = 0x68,
    UNKNOWN             = 0x70,
    MY_GAMES            = 0x80
};

struct SHITEM_UNKNOWN0 {
    uint16  size;
    uint8   type;
};

struct SHITEM_UNKNOWN1 {
    uint16  size;
    uint8   type;
};

struct SHITEM_ROOT_FOLDER {
    uint16          size;
    uint8           type;
    ROOTFOLDER_ID   folder_id;
    char            guid[16];
};

struct SHITEM_VOLUME {
    uint16  size;
    uint8   type;
};

struct SHITEM_FILE_ENTRY {
    uint16  size;
    uint8   type;
    uint8   unk0;
    uint32  file_size;
    uint32  modification_time;
    uint16  file_attribute_flags;
};

struct SHITEM_NETWORK {
    uint16  size;
    uint8   type;
    uint8   unk0;
    uint8   flags;
    char    location[];
};

struct SHITEM_COMPRESSED_FOLDER {
    uint16  size;
    uint8   type;
    uint8   unk0;
    uint16  unk1;
};

struct SHITEM_URI {
    uint16  size;
    uint8   type;
    uint8   flags;
    uint16  data_size;
};

struct SHITEM_CONTROL_PANEL {
    uint16  size;
    uint8   type;
    uint8   unk0;
    char    unk1[10];
    char    guid[16];
};

struct SHITEM_CONTROL_PANEL_CATEGORY {
    uint16  size;
    uint8   type;
    uint8   unk0;
    uint32  signature;
    uint32  category;
};

struct SHITEM_CDBURN {
    uint16  size;
    uint8   type;
    uint8   unk0;
    uint32  signature;
    uint32  unk1;
    uint32  unk2;
};

struct SHITEM_GAME_FOLDER {
    uint16  size;
    uint8   type;
    uint8   unk0;
    uint32  signature;
    char    identifier[16];
    uint64  unk1;
};

struct SHITEM_CONTROL_PANEL_CPL_FILE {
    uint16  size;
    uint8   type;
    uint8   unk0;
    uint32  signature;
    uint32  unk1;
    uint32  unk2;
    uint32  unk3;
    uint16  name_offset;
    uint16  comments_offset;
    wchar   cpl_path[];
    wchar   name[];
    wchar   comments[];
};

struct SHITEM_MTP_PROPERTY {
    char    format_identifier[16];
    uint32  value_identifier;
    uint32  value_type;
};

struct SHITEM_MTP_FILE_ENTRY {
    uint16  size;
    uint8   type;
    uint8   unk0;
    uint16  data_size;
    uint32  data_signature;
    uint32  unk1;
    uint16  unk2;
    uint16  unk3;
    uint16  unk4;
    uint16  unk5;
    uint32  unk6;
    uint64  modification_time;
    uint64  creation_time;
    char    content_type_folder[16];
    uint32  unk7;
    uint32  folder_name_size_1;
    uint32  folder_name_size_2;
    uint32  folder_identifier_size;
    wchar   folder_name_1[folder_name_size_1];
    wchar   folder_name_2[folder_name_size_2];
    uint32  unk8;
    char    class_identifier[16];
    uint32  num_properties;
};

struct SHITEM_MTP_VOLUME_GUID {
    wchar   guid[39];
};

struct SHITEM_MTP_VOLUME {
    uint16                      size;
    uint8                       type;
    uint8                       unk0;
    uint16                      data_size;
    uint32                      data_signature;
    uint32                      unk1;
    uint16                      unk2;
    uint16                      unk3;
    uint16                      unk4;
    uint16                      unk5;
    uint32                      unk6;
    uint64                      unk7;
    uint32                      unk8;
    uint32                      name_size;
    uint32                      identifier_size;
    uint32                      filesystem_size;
    uint32                      num_guid;
    wchar                       name[name_size];
    wchar                       identifier[identifier_size];
    wchar                       filesystem[filesystem_size];
    SHITEM_MTP_VOLUME_GUID      guids[num_guid];
    uint32                      unk9;
    char                        class_identifier[16];
    uint32                      num_properties;
};

struct SHITEM_USERS_PROPERTY_VIEW {
    uint16  size;
    uint8   type;
    uint8   unk0;
    uint16  data_size;
    uint32  data_signature;
    uint16  property_store_size;
    uint16  identifier_size;
    char    identifier[identifier_size];
    char    property_store[property_store_size];
    uint16  unk1;
};

struct SHITEM_UNKNOWN_0x74 {
    uint16  size;
    uint8   type;
    uint8   unk0;
    uint16  data_size;
    uint32  data_signature;
    uint16  subitem_size;
};

struct SHITEM_UNKNOWN_0x74_SUBITEM {
    uint8   type;
    uint8   unk1;
    uint32  file_size;
    uint32  modification_time;
    uint16  file_attribute_flags;
    char    primary_name[];
};

struct SHITEM_DELEGATE {
    uint16  size;
    uint8   type;
    uint8   unk0;
    uint16  data_size;
    char    data[data_size - 2];
    char    delegate_identifier[16];
    char    shell_identifier[16];
};

struct EXTENSION_BLOCK_HEADER {
    uint16  size;
    uint16  version;
    uint32  signature;
};
"""
c_bag = cstruct().load(bag_def)

DELEGATE_ITEM_IDENTIFIER = b"\x74\x1a\x59\x5e\x96\xdf\xd3\x48\x8d\x67\x17\x33\xbc\xee\x28\xba"


class SHITEM:
    STRUCT = None

    def __init__(self, buf: bytes):
        self.buf = buf
        self.fh = io.BytesIO(buf)
        self.item = self.STRUCT(self.fh) if self.STRUCT is not None else None
        self.size = self.item.size if self.item else len(self.buf)
        self.type = self.item.type if self.item else None
        self.parent = None
        self.extensions = []

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}>"

    @property
    def name(self) -> str:
        return f"<SHITEM 0x{self.size:x}>"

    @property
    def creation_time(self) -> None:
        return None

    @property
    def modification_time(self) -> None:
        return None

    @property
    def access_time(self) -> None:
        return None

    @property
    def file_size(self) -> None:
        return None

    @property
    def file_reference(self) -> None:
        return None

    def extension(self, cls: Any) -> Any | None:
        for ext in self.extensions:
            if isinstance(ext, cls):
                return ext
        return None


class UNKNOWN(SHITEM):
    @property
    def name(self) -> str:
        type_number = hex(self.type) if self.type else self.type
        return f"<UNKNOWN size=0x{self.size:04x} type={type_number}>"


class UNKNOWN0(SHITEM):
    STRUCT = c_bag.SHITEM_UNKNOWN0

    def __init__(self, buf: bytes):
        super().__init__(buf)
        self.guid = None

        if self.item.size == 0x20:
            self.guid = uuid.UUID(bytes_le=self.fh.read(16))

    @property
    def name(self) -> str:
        if self.guid:
            GUID_name = shell_folder_ids.DESCRIPTIONS.get(str(self.guid))
            return GUID_name or f"<UNKNOWN0: {{{self.guid}}}>"
        return f"<UNKNOWN0 0x{self.size:x}>"


class UNKNOWN1(SHITEM):
    STRUCT = c_bag.SHITEM_UNKNOWN1

    @property
    def name(self) -> str:
        return f"<UNKNOWN1 0x{self.size:x}>"


class ROOT_FOLDER(SHITEM):
    STRUCT = c_bag.SHITEM_ROOT_FOLDER

    def __init__(self, buf: bytes):
        super().__init__(buf)
        self.guid = uuid.UUID(bytes_le=self.item.guid)
        self.extension = None

        if self.item.size > 20:
            self.extension = None

    @property
    def name(self) -> str:
        GUID_name = shell_folder_ids.DESCRIPTIONS.get(str(self.guid))
        return GUID_name or f"{{{self.item.folder_id.name}: {self.guid}}}"


class VOLUME(SHITEM):
    STRUCT = c_bag.SHITEM_VOLUME

    def __init__(self, buf: bytes):
        super().__init__(buf)
        self.volume_name = None
        self.identifier = None
        if self.type == 0x2E:
            self.identifier = uuid.UUID(bytes_le=buf[4:20].tobytes())
        else:
            self.volume_name = self.fh.read(20).rstrip(b"\x00").decode(errors="surrogateescape")
            if self.size >= 41:
                self.identifier = uuid.UUID(bytes_le=buf[25:41].tobytes())

    @property
    def name(self) -> str:
        if self.volume_name:
            return self.volume_name
        if self.identifier:
            GUID_name = shell_folder_ids.DESCRIPTIONS.get(str(self.identifier))
            return GUID_name or f"{{{self.identifier}}}"
        return f"<VOLUME 0x{self.type:02x}>"


class FILE_ENTRY(SHITEM):
    STRUCT = c_bag.SHITEM_FILE_ENTRY

    def __init__(self, buf: bytes):
        super().__init__(buf)

        has_swn1 = False
        if buf[-30:] == b"S.W.N.1":
            has_swn1 = True

        if has_swn1 or self.type & 0x4:  # FILE_ENTRY_FLAG_IS_UNICODE
            self.primary_name = c_bag.wchar[None](self.fh)
            self.is_unicode = True
        else:
            self.primary_name = c_bag.char[None](self.fh).decode(errors="surrogateescape")
            self.is_unicode = False

        if self.fh.tell() % 2:
            self.fh.read(1)

        extension_size = c_bag.uint16(self.fh)
        self.fh.seek(-2, io.SEEK_CUR)

        self.is_pre_xp = False
        if not has_swn1 and ((self.size - self.fh.tell() < 2) or extension_size > self.size):
            self.is_pre_xp = True
            if self.is_unicode:
                self.secondary_name = c_bag.wchar[None](self.fh)
            else:
                self.secondary_name = c_bag.char[None](self.fh).decode(errors="surrogateescape")

    @property
    def name(self) -> str:
        ext = self.extension(EXTENSION_BLOCK_BEEF0004)
        if ext and ext.long_name:
            return ext.long_name
        return self.primary_name

    @property
    def modification_time(self) -> datetime | None:
        ts = self.item.modification_time
        if ts > 0:
            return dostimestamp(ts, swap=True).replace(tzinfo=timezone.utc)
        return None


class NETWORK(SHITEM):
    STRUCT = c_bag.SHITEM_NETWORK

    def __init__(self, buf: bytes):
        super().__init__(buf)
        self.description = None
        self.comments = None

        if self.item.flags & 0x80:
            self.description = c_bag.char[None](self.fh)

        if self.item.flags & 0x40:
            self.comments = c_bag.char[None](self.fh)

    @property
    def name(self) -> str:
        return self.item.location.decode(errors="surrogateescape")


class COMPRESSED_FOLDER(SHITEM):
    STRUCT = c_bag.SHITEM_COMPRESSED_FOLDER

    @property
    def name(self) -> str:
        return "<COMPRESSED_FOLDER>"


class URI(SHITEM):
    STRUCT = c_bag.SHITEM_URI

    def __init__(self, buf: bytes):
        super().__init__(buf)
        self.uri = None
        if self.item.data_size < self.size - 6:
            self.data = self.fh.read(self.item.data_size - 2)
            if self.item.flags & 0x80:
                self.uri = c_bag.wchar[None](self.fh)
            else:
                self.uri = c_bag.char[None](self.fh).decode(errors="surrogateescape")

    @property
    def name(self) -> str:
        return self.uri or "<URI>"


class CONTROL_PANEL(SHITEM):
    STRUCT = c_bag.SHITEM_CONTROL_PANEL

    def __init__(self, buf: bytes):
        super().__init__(buf)
        self.guid = uuid.UUID(bytes_le=self.item.guid)

    @property
    def name(self) -> str:
        GUID_name = shell_folder_ids.DESCRIPTIONS.get(str(self.guid))
        return GUID_name or f"<CONTROL_PANEL {self.guid}>"


class CONTROL_PANEL_CATEGORY(SHITEM):
    STRUCT = c_bag.SHITEM_CONTROL_PANEL_CATEGORY
    CATEGORIES: Final[dict[int, str]] = {
        0: "All Control Panel Items",
        1: "Appearance and Personalization",
        2: "Hardware and Sound",
        3: "Network and Internet",
        4: "Sounds, Speech, and Audio Devices",
        5: "System and Security",
        6: "Clock, Language, and Region",
        7: "Ease of Access",
        8: "Programs",
        9: "User Accounts",
        10: "Security Center",
        11: "Mobile PC",
    }

    @property
    def name(self) -> str:
        categ_str = self.CATEGORIES.get(self.item.category)
        if categ_str:
            return categ_str
        return f"<CONTROL_PANEL_CATEGORY 0x{self.item.unk1:08x}>"


class CDBURN(SHITEM):
    STRUCT = c_bag.SHITEM_CDBURN

    @property
    def name(self) -> str:
        return "<CDBURN>"


class GAME_FOLDER(SHITEM):
    STRUCT = c_bag.SHITEM_GAME_FOLDER

    def __init__(self, buf: bytes):
        super().__init__(buf)
        self.guid = uuid.UUID(bytes_le=self.item.identifier)

    @property
    def name(self) -> str:
        return f"<GAME_FOLDER {{{self.guid}}}>"


class CONTROL_PANEL_CPL_FILE(SHITEM):
    STRUCT = c_bag.SHITEM_CONTROL_PANEL_CPL_FILE

    @property
    def name(self) -> str:
        return f"<CONTROL_PANEL_CPL_FILE path={self.item.cpl_path} name={self.item.name} comments={self.item.comments}>"


class MTP_FILE_ENTRY(SHITEM):
    STRUCT = c_bag.SHITEM_MTP_FILE_ENTRY

    @property
    def name(self) -> str:
        return "<MTP_FILE_ENTRY>"

    @property
    def creation_time(self) -> datetime:
        return self.item.creation_time

    @property
    def modification_time(self) -> datetime:
        return self.item.modification_time


class MTP_VOLUME(SHITEM):
    STRUCT = c_bag.SHITEM_MTP_FILE_ENTRY

    @property
    def name(self) -> str:
        return "<MTP_VOLUME>"


class USERS_PROPERTY_VIEW(SHITEM):
    STRUCT = c_bag.SHITEM_USERS_PROPERTY_VIEW

    def __init__(self, buf: bytes):
        super().__init__(buf)
        self.guid = None
        self.identifier = self.item.data_signature

        if self.item.identifier_size == 16:
            self.guid = uuid.UUID(bytes_le=self.item.identifier)

    @property
    def name(self) -> str:
        # As we don't know how to handle identifier_size other than 16 bytes, we fall back to data_signature
        property_view = self.guid or self.identifier
        return f"<USERS_PROPERTY_VIEW {{{property_view}}}>"


class UNKNOWN_0x74(SHITEM):
    STRUCT = c_bag.SHITEM_UNKNOWN_0x74

    def __init__(self, buf: bytes):
        super().__init__(buf)
        self.subitem = None
        if self.item.subitem_size >= 16:
            self.subitem = c_bag.SHITEM_UNKNOWN_0x74_SUBITEM(self.fh)

    @property
    def name(self) -> str:
        return self.subitem.primary_name.decode(errors="surrogateescape") if self.subitem else "<UNKNOWN_0x74>"

    @property
    def modification_time(self) -> datetime | None:
        if self.subitem.modification_time > 0:
            return (
                dostimestamp(self.subitem.modification_time, swap=True).replace(tzinfo=timezone.utc)
                if self.subitem
                else None
            )
        return None


class DELEGATE(SHITEM):
    STRUCT = c_bag.SHITEM_DELEGATE

    def __init__(self, buf: bytes):
        super().__init__(buf)
        self.delegate_identifier = uuid.UUID(bytes_le=self.item.delegate_identifier)
        self.shell_identifier = uuid.UUID(bytes_le=self.item.shell_identifier)

    @property
    def name(self) -> str:
        GUID_name = shell_folder_ids.DESCRIPTIONS.get(str(self.shell_identifier))
        return GUID_name if GUID_name else f"{{{self.shell_identifier}}}"


class EXTENSION_BLOCK:
    def __init__(self, buf: bytes):
        self.buf = buf
        self.fh = io.BytesIO(buf)
        self.header = c_bag.EXTENSION_BLOCK_HEADER(self.fh)

    def __repr__(self) -> str:
        return f"<EXTENSION_BLOCK size=0x{self.size:04x} version=0x{self.version:04x} signature=0x{self.signature:08x}>"

    @property
    def size(self) -> int:
        return self.header.size

    @property
    def data_size(self) -> int:
        return self.size - 8  # minus header

    @property
    def version(self) -> int:
        return self.header.version

    @property
    def signature(self) -> int:
        return self.header.signature


class EXTENSION_BLOCK_BEEF0004(EXTENSION_BLOCK):
    def __init__(self, buf: bytes):
        super().__init__(buf)
        fh = self.fh
        version = self.version
        self.creation_time = c_bag.uint32(fh)
        self.last_accessed = c_bag.uint32(fh)
        self.identifier = c_bag.uint16(fh)
        self.file_reference = None
        self.long_name = None
        self.localized_name = None
        # Note that the c_bag.uintXX() etc. statements advance the pointer into
        # the filebuffer, so the order of the if statements is important here.
        if version >= 7:
            c_bag.uint16(fh)
            self.file_reference = c_bag.uint64(fh)
            c_bag.uint64(fh)
        if version >= 3:
            # Start of strings
            localized_name_offset = c_bag.uint16(fh)
        if version >= 9:
            c_bag.uint32(fh)
        if version >= 8:
            c_bag.uint32(fh)
        if version >= 3:
            self.long_name = c_bag.wchar[None](fh)

        if 3 <= version < 7 and localized_name_offset > 0:
            self.localized_name = c_bag.char[None](fh)

        if version >= 7 and localized_name_offset > 0:
            self.localized_name = c_bag.wchar[None](fh)


class EXTENSION_BLOCK_BEEF0005(EXTENSION_BLOCK):
    def __init__(self, buf: bytes):
        super().__init__(buf)
        c_bag.char[16](self.fh)  # GUID?
        self.shell_items = self.fh.read(self.data_size - 18)


ShellBagRecord = create_extended_descriptor([RegistryRecordDescriptorExtension, UserRecordDescriptorExtension])(
    "windows/shellbag",
    [
        ("datetime", "ts_mtime"),
        ("datetime", "ts_atime"),
        ("datetime", "ts_btime"),
        ("string", "type"),
        ("path", "path"),
        ("datetime", "regf_mtime"),
    ],
)


class ShellBagsPlugin(Plugin):
    """Windows Shellbags plugin."""

    KEYS = (
        "HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\Shell",
        "HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\ShellNoRoam",
        "HKEY_CURRENT_USER\\Software\\Classes\\Local Settings\\Software\\Microsoft\\Windows\\Shell",
        "HKEY_CURRENT_USER\\Software\\Classes\\Local Settings\\Software\\Microsoft\\Windows\\ShellNoRoam",
        "HKEY_CURRENT_USER\\Software\\Classes\\Wow6432Node\\Local Settings\\Software\\Microsoft\\Windows\\Shell",
        "HKEY_CURRENT_USER\\Software\\Classes\\Wow6432Node\\Local Settings\\Software\\Microsoft\\Windows\\ShellNoRoam",
        "HKEY_CURRENT_USER\\Local Settings\\Software\\Microsoft\\Windows\\Shell\\BagMRU",
    )

    def __init__(self, target: Target):
        super().__init__(target)
        self.bagkeys: list[RegistryKey] = list(self.target.registry.keys(self.KEYS))

    def check_compatible(self) -> None:
        if not self.bagkeys:
            raise UnsupportedPluginError("No shellbags found")

    @export(record=ShellBagRecord)
    def shellbags(self) -> Iterator[ShellBagRecord]:
        """Yields Windows Shellbags.

        Shellbags are registry keys to improve user experience when using Windows Explorer.
        They contain information such as file and folder creation time and access time.

        References:
            - https://github.com/libyal/libfwsi
            - https://www.giac.org/paper/gcfa/9576/windows-shellbag-forensics-in-depth/128522
            - https://www.hackingarticles.in/forensic-investigation-shellbags/
        """
        for regkey in self.bagkeys:
            try:
                yield from self._walk_bags(regkey.subkey("BagMRU"), None)

            except RegistryKeyNotFoundError:  # noqa: PERF203
                continue

            except Exception as e:
                self.target.log.error("Exception while parsing shellbags")  # noqa: TRY400
                self.target.log.debug("", exc_info=e)
                continue

    def _walk_bags(self, key: RegistryKey, path_prefix: str | None) -> Iterator[ShellBagRecord]:
        """Recursively walk shellbags from the given RegistryKey location."""
        path_prefix = [] if path_prefix is None else [path_prefix]
        user = self.target.registry.get_user(key)

        for reg_val in key.values():
            name, value = reg_val.name, reg_val.value
            if not name.isdigit():
                continue
            path = None

            for item in parse_shell_item_list(value):
                path = "\\".join([*path_prefix, item.name])
                yield ShellBagRecord(
                    ts_mtime=item.modification_time,
                    ts_atime=item.access_time,
                    ts_btime=item.creation_time,
                    type=item.__class__.__name__,
                    path=windows_path(path),
                    regf_mtime=key.ts,
                    _key=key,
                    _user=user,
                    _target=self.target,
                )

            yield from self._walk_bags(key.subkey(name), path)


def parse_shell_item_list(buf: bytes) -> Iterator[SHITEM]:
    """Parse a shellbag item from the given bytes."""

    offset = 0
    end = len(buf)
    list_buf = memoryview(buf)
    parent = None

    while offset < end:
        size = c_bag.uint16(list_buf[offset : offset + 2])

        if size == 0:
            break

        item_buf = list_buf[offset : offset + size]

        entry = None
        if size >= 8:
            signature = c_bag.uint32(item_buf[4:8])
            if signature == 0x39DE2184:
                entry = CONTROL_PANEL_CATEGORY
            elif signature == 0x4D677541:
                entry = CDBURN
            elif signature == 0x49534647:
                entry = GAME_FOLDER
            elif signature == 0xFFFFFF38:
                entry = CONTROL_PANEL_CPL_FILE

        if size >= 10 and not entry:
            signature = c_bag.uint32(item_buf[6:10])
            if signature == 0x07192006:
                entry = MTP_FILE_ENTRY
            elif signature == 0x10312005:
                entry = MTP_VOLUME
            elif signature in (0x10141981, 0x23A3DFD5, 0x23FEBBEE, 0x3B93AFBB, 0xBEEBEE00):
                entry = USERS_PROPERTY_VIEW
            elif signature == 0x46534643:
                entry = UNKNOWN_0x74

        if size >= 38 and not entry and item_buf[size - 32 : size] == DELEGATE_ITEM_IDENTIFIER:
            entry = DELEGATE

        if size >= 3 and not entry:
            class_type = item_buf[2]
            mask_type = class_type & 0x70

            if mask_type == 0x00:
                if class_type == 0x00:
                    entry = UNKNOWN0
                elif class_type == 0x01:
                    entry = UNKNOWN1

            elif mask_type == 0x10:
                if class_type == 0x1F:
                    entry = ROOT_FOLDER

            elif mask_type == 0x20:
                if class_type in (0x23, 0x25, 0x29, 0x2A, 0x2E, 0x2F):
                    entry = VOLUME

            elif mask_type == 0x30:
                if class_type in (0x30, 0x31, 0x32, 0x35, 0x36, 0xB1):
                    entry = FILE_ENTRY

            elif mask_type == 0x40:
                if class_type in (0x41, 0x42, 0x46, 0x47, 0x4C, 0xC3):
                    entry = NETWORK

            elif mask_type == 0x50:
                if class_type == 0x52:
                    entry = COMPRESSED_FOLDER

            elif mask_type == 0x60:
                if class_type == 0x61:
                    entry = URI

            elif mask_type == 0x70:
                if class_type == 0x71:
                    entry = CONTROL_PANEL
            else:
                if not entry:
                    log.debug("No supported shell item found for size 0x%04x and type 0x%02x", size, class_type)
                    entry = UNKNOWN

        if not entry:
            log.debug("No supported shell item found for size 0x%04x", size)
            entry = UNKNOWN

        entry = entry(item_buf)
        entry.parent = parent

        first_extension_block_offset = c_bag.uint16(item_buf[-2:])
        if 4 <= first_extension_block_offset < size - 2:
            extension_offset = first_extension_block_offset
            while extension_offset < size - 2:
                extension_size = c_bag.uint16(item_buf[extension_offset : extension_offset + 2])

                if extension_size == 0:
                    break

                if extension_size > size - extension_offset:
                    log.debug(
                        "Extension size exceeds item size: 0x%04x > 0x%04x - 0x%04x",
                        extension_size,
                        size,
                        extension_offset,
                    )
                    break  # Extension size too large

                extension_buf = item_buf[extension_offset : extension_offset + extension_size]
                extension_signature = c_bag.uint32(extension_buf[4:8])

                ext = None

                if extension_signature >> 16 != 0xBEEF:
                    log.debug("Got unsupported extension signature 0x%08x from item %r", extension_signature, entry)
                    # Unsupported

                elif extension_signature == 0xBEEF0000 or extension_signature == 0xBEEF0001:
                    pass

                elif extension_signature == 0xBEEF0003 or extension_signature == 0xBEEF0004:
                    ext = EXTENSION_BLOCK_BEEF0004

                elif extension_signature == 0xBEEF0005:
                    ext = EXTENSION_BLOCK_BEEF0005

                elif extension_signature in (
                    0xBEEF0006,
                    0xBEEF000A,
                    0xBEEF0013,
                    0xBEEF0014,
                    0xBEEF0019,
                    0xBEEF0025,
                    0xBEEF0026,
                ):
                    pass

                else:
                    log.debug(
                        "Got unsupported beef extension signature 0x%08x from item %r", extension_signature, entry
                    )

                if ext is None:
                    ext = EXTENSION_BLOCK
                    log.debug("Unimplemented extension signature 0x%08x from item %r", extension_signature, entry)

                ext = ext(extension_buf)

                entry.extensions.append(ext)
                extension_offset += extension_size

        parent = entry
        offset += size
        yield entry



from __future__ import annotations

from typing import TYPE_CHECKING, BinaryIO

import dissect.qnxfs as qnxfs
from dissect.qnxfs.qnx4 import INode4
from dissect.qnxfs.qnx6 import INode6

from dissect.target.exceptions import (
    FileNotFoundError,
    FilesystemError,
    IsADirectoryError,
    NotADirectoryError,
    NotASymlinkError,
)
from dissect.target.filesystem import DirEntry, Filesystem, FilesystemEntry
from dissect.target.helpers import fsutil

if TYPE_CHECKING:
    from collections.abc import Iterator

INode = INode4 | INode6


class QnxFilesystem(Filesystem):
    __type__ = "qnxfs"

    def __init__(self, fh: BinaryIO, *args, **kwargs):
        super().__init__(fh, *args, **kwargs)
        self.qnxfs = qnxfs.QNXFS(fh)

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        return qnxfs.is_qnxfs(fh)

    def get(self, path: str) -> FilesystemEntry:
        return QnxFilesystemEntry(self, path, self._get_node(path))

    def _get_node(self, path: str, node: INode | None = None) -> INode:
        try:
            return self.qnxfs.get(path, node)
        except qnxfs.FileNotFoundError as e:
            raise FileNotFoundError(path) from e
        except qnxfs.NotADirectoryError as e:
            raise NotADirectoryError(path, cause=e)
        except qnxfs.NotASymlinkError as e:
            raise NotASymlinkError(path, cause=e)
        except qnxfs.Error as e:
            raise FileNotFoundError(path, cause=e)


class QnxDirEntry(DirEntry):
    fs: QnxFilesystem
    entry: INode

    def get(self) -> QnxFilesystemEntry:
        return QnxFilesystemEntry(self.fs, self.path, self.entry)

    def stat(self, *, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self.get().stat(follow_symlinks=follow_symlinks)


class QnxFilesystemEntry(FilesystemEntry):
    fs: QnxFilesystem
    entry: INode

    def get(self, path: str) -> FilesystemEntry:
        entry_path = fsutil.join(self.path, path, alt_separator=self.fs.alt_separator)
        return QnxFilesystemEntry(self.fs, entry_path, self.fs._get_node(path, self.entry))

    def open(self) -> BinaryIO:
        if self.is_dir():
            raise IsADirectoryError(self.path)
        return self._resolve().entry.open()

    def scandir(self) -> Iterator[QnxDirEntry]:
        if not self.is_dir():
            raise NotADirectoryError(self.path)

        for name, entry in self._resolve().entry.iterdir():
            # TODO: Separate INode and DirEntry in dissect.qnxfs
            yield QnxDirEntry(self.fs, self.path, name, entry)

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        try:
            return self._resolve(follow_symlinks).entry.is_dir()
        except FilesystemError:
            return False

    def is_file(self, follow_symlinks: bool = True) -> bool:
        try:
            return self._resolve(follow_symlinks).entry.is_file()
        except FilesystemError:
            return False

    def is_symlink(self) -> bool:
        return self.entry.is_symlink()

    def readlink(self) -> str:
        if not self.is_symlink():
            raise NotASymlinkError

        return self.entry.link

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self._resolve(follow_symlinks).lstat()

    def lstat(self) -> fsutil.stat_result:
        # mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime
        st_info = fsutil.stat_result(
            [
                self.entry.mode,
                self.entry.inum,
                id(self.fs),
                getattr(self.entry, "nlink", 1),  # Only QNX4 has nlink
                self.entry.uid,
                self.entry.gid,
                self.entry.size,
                self.entry.atime.timestamp(),
                self.entry.mtime.timestamp(),
                self.entry.ctime.timestamp(),
            ]
        )

        st_info.st_birthtime = self.entry.ftime.timestamp()

        st_info.st_blksize = self.fs.qnxfs.block_size
        # Blocks are always calculated based on 512 byte blocks
        # https://github.com/torvalds/linux/blob/c1e939a21eb111a6d6067b38e8e04b8809b64c4e/fs/qnx6/inode.c#L560-L561
        st_info.st_blocks = (self.entry.size + 511) >> 9

        return st_info



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugin import internal
from dissect.target.plugins.os.unix.bsd._os import BsdPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target.filesystem import Filesystem
    from dissect.target.target import Target

# https://en.wikipedia.org/wiki/Mach-O
ARCH_MAP = {
    b"\x0c\x00\x00\x01": "arm64",  # big endian, x64
    b"\x01\x00\x00\x0c": "arm64",  # little endian, x64
    b"\x0c\x00\x00\x00": "arm32",  # big endian, x32
    b"\x00\x00\x00\x0c": "arm32",  # little endian, x32
}


class DarwinPlugin(BsdPlugin):
    """Darwin plugin."""

    def __init__(self, target: Target):
        super().__init__(target)

    @classmethod
    def detect(cls, target: Target) -> Filesystem | None:
        for fs in target.filesystems:
            if (fs.exists("/Library") and fs.exists("/Applications")) or fs.exists("/private/var/mobile"):
                return fs
        return None

    @internal
    def misc_user_paths(self) -> Iterator[tuple[str, tuple[str, str] | None]]:
        yield from super().misc_user_paths()

        if (user_path := self.target.fs.path("/Users")).exists():
            yield from ((entry, None) for entry in user_path.iterdir() if entry.is_dir())


def detect_macho_arch(paths: list[str | Path], fs: Filesystem | None = None) -> str | None:
    """Detect the architecture of the system by reading the Mach-O headers of the provided binaries.

    We could use the mach-o magic headers (feedface, feedfacf, cafebabe), but the mach-o cpu type
    also contains bitness.

    Args:
        paths: List of strings or ``Path`` objects.
        fs: Optional filesystem to search the provided paths in. Required if ``paths`` is a list of strings.

    Returns:
        Detected architecture (e.g. ``arm64``) or ``None``.

    References:
        - https://github.com/opensource-apple/cctools/blob/master/include/mach/machine.h
    """
    for path in paths:
        if isinstance(path, str):
            if not fs:
                raise ValueError("Provided string paths but no filesystem!")
            path = fs.path(path)

        if not path.is_file():
            continue

        try:
            with path.open("rb") as fh:
                fh.seek(4)
                return ARCH_MAP.get(fh.read(4))  # mach-o cpu type
        except Exception:
            pass

    return None



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.util import ts

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from datetime import datetime


class GenericPlugin(Plugin):
    """Generic Citrix plugin."""

    def check_compatible(self) -> None:
        if self.target.os != "citrix-netscaler":
            raise UnsupportedPluginError("Citrix Netscaler specific plugin loaded on non-Citrix target")

    @export(property=True)
    def install_date(self) -> datetime | None:
        """Return the likely install date of Citrix Netscaler."""
        entries = ["/flash/.version", "/flash", "/var"]
        for entry in entries:
            if (path := self.target.fs.path(entry)).exists():
                # Prioritize birth time if available
                stat = path.stat()
                return ts.from_unix(stat.st_birthtime or stat.st_ctime)
        return None



from __future__ import annotations

from typing import TYPE_CHECKING, BinaryIO
from unittest.mock import patch

import pytest

from dissect.target import volume
from dissect.target.helpers import keychain
from tests._utils import absolute_path

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

try:
    from dissect.target.volumes.bde import BitlockerVolumeSystem

    HAS_DISSECT_FVE = True
except ModuleNotFoundError:
    HAS_DISSECT_FVE = False


@pytest.fixture
def encrypted_volume() -> Iterator[BinaryIO]:
    data_file = "_data/volumes/bde/enc-volume.bin"
    with absolute_path(data_file).open("rb") as fh:
        yield fh


@pytest.mark.skipif(not HAS_DISSECT_FVE, reason="requires dissect.fve")
def test_bde_volume_failure(target_win: Target, encrypted_volume: BinaryIO) -> None:
    """Test if not providing a key to decrypt a BDE volume results in no mounted volume."""

    enc_vol = volume.Volume(encrypted_volume, 1, 0, None, None, None, disk=encrypted_volume)
    target_win.volumes.add(enc_vol)
    target_win.volumes.apply()

    assert len(target_win.volumes) == 1
    assert enc_vol in target_win.volumes


@pytest.mark.skipif(not HAS_DISSECT_FVE, reason="requires dissect.fve")
def test_bde_volume_with_recovery_key(target_win: Target, encrypted_volume: BinaryIO) -> None:
    """Test if we can decrypt a BDE volume using a recovery key."""

    recovery_key = "272316-265804-640728-713570-509047-503305-045837-324731"

    with patch.object(keychain, "KEYCHAIN", []):
        keychain.register_key(
            keychain.KeyType.RECOVERY_KEY,
            recovery_key,
            identifier=None,
            provider=BitlockerVolumeSystem.__type__,
        )

        enc_vol = volume.Volume(encrypted_volume, 1, 0, None, None, None, disk=encrypted_volume)
        target_win.volumes.add(enc_vol)
        target_win.volumes.apply()

        assert len(target_win.volumes) == 2
        assert enc_vol in target_win.volumes

        dec_vol = next(v for v in target_win.volumes if v != enc_vol)

        # virtual fs + ntfs fs
        assert len(target_win.filesystems) == 2
        target_win.fs.mount("e:/", dec_vol.fs)

        assert target_win.fs.path("e:/test-folder/test-file-2.txt").exists()


@pytest.mark.skipif(not HAS_DISSECT_FVE, reason="requires dissect.fve")
def test_bde_volume_with_passphrase(target_win: Target, encrypted_volume: BinaryIO) -> None:
    """Test if we can decrypt a BDE volume using a passphrase."""

    identifier = "B6AD258A-2725-4A42-93C6-844478BF7A90"
    passphrase = "Password1234"

    with patch.object(keychain, "KEYCHAIN", []):
        keychain.register_key(
            keychain.KeyType.PASSPHRASE,
            passphrase,
            identifier=identifier,
            provider=BitlockerVolumeSystem.__type__,
        )

        enc_vol = volume.Volume(encrypted_volume, 1, 0, None, None, None, disk=encrypted_volume)
        target_win.volumes.add(enc_vol)
        target_win.volumes.apply()

        assert len(target_win.volumes) == 2
        assert enc_vol in target_win.volumes

        dec_vol = next(v for v in target_win.volumes if v != enc_vol)

        # virtual fs + ntfs fs
        assert len(target_win.filesystems) == 2
        target_win.fs.mount("e:/", dec_vol.fs)

        assert target_win.fs.path("e:/test-folder/test-file-2.txt").exists()


@pytest.mark.skipif(not HAS_DISSECT_FVE, reason="requires dissect.fve")
def test_bde_volume_with_wildcard_key(target_win: Target, encrypted_volume: BinaryIO) -> None:
    """Test if we can decrypt a BDE volume using a wildcard keychain key."""

    with patch.object(keychain, "KEYCHAIN", []):
        keychain.register_wildcard_value("Password1234")

        enc_vol = volume.Volume(encrypted_volume, 1, 0, None, None, None, disk=encrypted_volume)
        target_win.volumes.add(enc_vol)
        target_win.volumes.apply()

        assert len(target_win.volumes) == 2
        assert enc_vol in target_win.volumes

        dec_vol = next(v for v in target_win.volumes if v != enc_vol)

        # virtual fs + ntfs fs
        assert len(target_win.filesystems) == 2
        target_win.fs.mount("e:/", dec_vol.fs)

        assert target_win.fs.path("e:/test-folder/test-file-2.txt").exists()


@pytest.mark.skipif(not HAS_DISSECT_FVE, reason="requires dissect.fve")
def test_bde_volume_with_raw_key(target_win: Target, encrypted_volume: BinaryIO) -> None:
    """Test if we can decrypt a BDE volume using a raw FVEK key."""

    with patch.object(keychain, "KEYCHAIN", []):
        keychain.register_wildcard_value("ab60a58f1a0b60be91ffa2b40ec338a072a011302a8ff58fc45eee742711dc7f")

        enc_vol = volume.Volume(encrypted_volume, 1, 0, None, None, None, disk=encrypted_volume)
        target_win.volumes.add(enc_vol)
        target_win.volumes.apply()

        assert len(target_win.volumes) == 2
        assert enc_vol in target_win.volumes

        dec_vol = next(v for v in target_win.volumes if v != enc_vol)

        # virtual fs + ntfs fs
        assert len(target_win.filesystems) == 2
        target_win.fs.mount("e:/", dec_vol.fs)

        assert target_win.fs.path("e:/test-folder/test-file-2.txt").exists()



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.clfs import blf, container
from dissect.clfs.exceptions import InvalidBLFError, InvalidRecordBlockError

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

ClfsRecord = TargetRecordDescriptor(
    "filesystem/windows/clfs",
    [
        ("string", "stream_name"),
        ("string", "type"),
        ("string", "file_attributes"),
        ("uint32", "offset"),
        ("string", "container_name"),
        ("uint32", "container_size"),
        ("uint32", "record_offset"),
        ("bytes", "record_data"),
        ("bytes", "block_data"),
        ("uint32", "clfs_stream_id"),
        ("uint32", "clfs_container_id"),
    ],
)


class ClfsPlugin(Plugin):
    """CLFS Plugin.

    Dissect plugin for parsing the Base Log Files of a Microsoft Windows system.

    Most of these records are actually parsed in-memory, this is the first iteration
    to parse the files present on disk. This should be improved in the near future when
    the memory implementation for dissect is working.
    """

    BLF_PATH = "%windir%/system32/config/"  # Unsure at time of writing if this is the only location

    def __init__(self, target: Target):
        super().__init__(target)
        self._blfs = []

        blfdir = self.target.resolve(self.BLF_PATH)

        if blfdir.exists() and blfdir.is_dir():
            blf_files = blfdir.glob("*.blf")

            for blf_path in blf_files:
                fh = blf_path.open()

                try:
                    blf_instance = blf.BLF(fh)
                    self._blfs.append((blf_path, blf_instance))
                except InvalidRecordBlockError as e:
                    self.target.log.warning("Invalid record block: %s", blf_path)
                    self.target.log.debug("", exc_info=e)
                except InvalidBLFError as e:
                    self.target.log.warning("Could not validate BLF: %s", blf_path)
                    self.target.log.debug("", exc_info=e)

    def check_compatible(self) -> None:
        if not self._blfs:
            raise UnsupportedPluginError("No BLF files found")

    @export(record=ClfsRecord)
    def clfs(self) -> Iterator[ClfsRecord]:
        """Parse the containers associated with a valid BLF file.

        Containers are used to store the transactional logs in the form of records.

        References:
            - https://docs.microsoft.com/en-us/windows-hardware/drivers/kernel/introduction-to-the-common-log-file-system
        """

        for blf_path, blf_instance in self._blfs:
            # We only parse the base record client/container contexts for now
            for base_record in blf_instance.base_records():
                for stream in base_record.streams:
                    for blf_container in base_record.containers:
                        # Check if the stream ID is matching the container ID
                        if blf_container.id != stream.lsn_base.Offset.ContainerId:
                            continue

                        # We can encounter the same container ID for the shadow blocks
                        if blf_container.type != stream.type:
                            continue

                        # Invalid LSN (-1)
                        if stream.lsn_base.PhysicalOffset <= 0:
                            continue

                        container_path = blf_container.name.replace("%BLF%", str(blf_path.parent))
                        container_file = self.target.fs.path(container_path)

                        fh = container_file.open()
                        trans = container.Container(fh, offset=stream.offset)

                        # Open each container and yield the results for each record found within that container
                        for record_offset, record_data, block_data in trans.records():
                            yield ClfsRecord(
                                stream_name=stream.name,
                                type=stream.type,
                                file_attributes=stream.file_attributes,
                                offset=stream.offset,
                                container_name=container_file.name,
                                container_size=blf_container.size,
                                record_offset=record_offset,
                                record_data=record_data,
                                block_data=block_data,
                                clfs_stream_id=stream.id,
                                clfs_container_id=stream.lsn_base.Offset.ContainerId,
                                _target=self.target,
                            )



from __future__ import annotations

import subprocess
from pathlib import Path

import pytest

try:
    subprocess.run(["git", "lfs", "--version"], check=True)
    HAS_GIT_LFS = True
except Exception:
    HAS_GIT_LFS = False


@pytest.mark.skipif(not HAS_GIT_LFS, reason="git lfs command not available")
def test_lfs_tracking(request: pytest.FixtureRequest) -> None:
    """Test if all files that should be checked into git lfs are actually tracked."""
    output = subprocess.run(["git", "lfs", "ls-files", "--name-only"], capture_output=True, text=True)
    tracked_lfs_files = set(output.stdout.splitlines())

    output = subprocess.run(["git", "ls-files"], capture_output=True, text=True)
    tracked_git_files = set(output.stdout.splitlines())

    with (request.config.rootdir / ".gitattributes").open("rt") as fh:
        lfs_patterns = [line.split()[0] for line in fh.readlines() if "filter=lfs" in line]

    untracked_files = []
    for pattern in lfs_patterns:
        if pattern.endswith("/**"):
            pattern += "/*"

        for path in Path(request.config.rootdir).rglob(pattern):
            relative_path = str(path.relative_to(request.config.rootdir))
            if (
                path.stat().st_size != 0
                and relative_path in tracked_git_files
                and relative_path not in tracked_lfs_files
            ):
                untracked_files.append(relative_path)

    assert not untracked_files, f"Untracked LFS files: {', '.join(untracked_files)}"



from __future__ import annotations

import io
from io import BytesIO
from typing import TYPE_CHECKING, BinaryIO

import pytest

from dissect.target import container
from dissect.target.containers.split import SplitContainer

if TYPE_CHECKING:
    from pathlib import Path


def _assert_split_container(fh: SplitContainer) -> None:
    assert isinstance(fh, SplitContainer)


@pytest.fixture
def split_fhs() -> list[BinaryIO]:
    return [BytesIO(bytes([char] * 512)) for char in b"ABCD"]


@pytest.fixture
def split_paths(tmp_path: Path, split_fhs: list[BinaryIO]) -> list[Path]:
    paths = [(tmp_path / f"split.{i:>03}") for i in range(4)]

    for fh, path in zip(split_fhs, paths, strict=False):
        fh.seek(0)
        path.write_bytes(fh.read())

    return paths


@pytest.fixture
def split_symlink(tmp_path: Path, split_paths: list[Path]) -> Path:
    dir_path = tmp_path / "dir"
    symlink_path = dir_path / split_paths[0].name
    dir_path.mkdir()
    symlink_path.symlink_to(split_paths[0])

    return symlink_path


@pytest.mark.parametrize(
    "obj",
    [
        "split_fhs",
        "split_paths",
        "split_symlink",
    ],
)
def test_split_container(obj: str, request: pytest.FixtureRequest) -> None:
    fh = container.open(request.getfixturevalue(obj))
    _assert_split_container(fh)
    assert fh.read(4096) == (b"A" * 512) + (b"B" * 512) + (b"C" * 512) + (b"D" * 512)
    fh.seek(0, whence=io.SEEK_SET)
    assert fh.read(4) == b"AAAA"
    assert fh.tell() == 4
    fh.seek(508, whence=io.SEEK_CUR)
    assert fh.read(4) == b"BBBB"
    fh.seek(-4, whence=io.SEEK_END)
    assert fh.read(4) == b"DDDD"



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import ChildTargetRecord
from dissect.target.plugin import ChildTargetPlugin
from dissect.target.plugins.apps.virtualization.vmware_workstation import find_vm_inventory, parse_inventory_file

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target


class VmwareWorkstationChildTargetPlugin(ChildTargetPlugin):
    """Child target plugin that yields from VMware Workstation VM inventory."""

    __type__ = "vmware_workstation"

    def __init__(self, target: Target):
        super().__init__(target)
        self.inventories = [inventory for inventory, _ in find_vm_inventory(target)]

    def check_compatible(self) -> None:
        if not self.inventories:
            raise UnsupportedPluginError("No VMWare Workstation inventories found")

    def list_children(self) -> Iterator[ChildTargetRecord]:
        for inv in self.inventories:
            inventory = parse_inventory_file(inv)

            for config in inventory.values():
                yield ChildTargetRecord(
                    type=self.__type__,
                    name=config.get("DisplayName"),
                    path=config.get("config"),
                    _target=self.target,
                )



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.filesystem import Filesystem
from dissect.target.plugin import OperatingSystem, export
from dissect.target.plugins.os.unix._os import UnixPlugin

if TYPE_CHECKING:
    from dissect.target.filesystem import Filesystem
    from dissect.target.target import Target


class BsdPlugin(UnixPlugin):
    def __init__(self, target: Target):
        super().__init__(target)

    @classmethod
    def detect(cls, target: Target) -> Filesystem | None:
        for fs in target.filesystems:
            # checking the existence of /var/authpf for free- and openbsd
            # checking the existence of /var/at for net- and freebsd
            if fs.exists("/usr/obj") or fs.exists("/altroot") or fs.exists("/etc/pf.conf"):
                return fs

        return None

    @export(property=True)
    def os(self) -> str:
        return OperatingSystem.BSD.value

    @export(property=True)
    def hostname(self) -> str | None:
        for name in ("/etc/rc.conf", "/etc/rc.conf.defaults"):
            if (file := self.target.fs.path(name)).is_file():
                with file.open("rt") as fh:
                    for line in fh:
                        if line.startswith("hostname"):
                            return line.rstrip().split("=", maxsplit=1)[1].replace('"', "")

        return super().hostname

    @export(property=True)
    def ips(self) -> list[str] | None:
        self.target.log.error("ips plugin not implemented for %s", self.__class__)
        return None
